!-------------------------------------------------------------------------------

!                      Code_Saturne version 2.0.0
!                      --------------------------

!     This file is part of the Code_Saturne Kernel, element of the
!     Code_Saturne CFD tool.

!     Copyright (C) 1998-2009 EDF S.A., France

!     contact: saturne-support@edf.fr

!     The Code_Saturne Kernel is free software; you can redistribute it
!     and/or modify it under the terms of the GNU General Public License
!     as published by the Free Software Foundation; either version 2 of
!     the License, or (at your option) any later version.

!     The Code_Saturne Kernel is distributed in the hope that it will be
!     useful, but WITHOUT ANY WARRANTY; without even the implied warranty
!     of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!     GNU General Public License for more details.

!     You should have received a copy of the GNU General Public License
!     along with the Code_Saturne Kernel; if not, write to the
!     Free Software Foundation, Inc.,
!     51 Franklin St, Fifth Floor,
!     Boston, MA  02110-1301  USA

!-------------------------------------------------------------------------------

subroutine usvpst &
!================

 ( idbia0 , idbra0 , ipart  ,                                     &
   ndim   , ncelet , ncel   , nfac   , nfabor , nfml   , nprfml , &
   nnod   , lndfac , lndfbr , ncelbr ,                            &
   nvar   , nscal  , nphas  , nvlsta ,                            &
   ncelps , nfacps , nfbrps ,                                     &
   nideve , nrdeve , nituse , nrtuse ,                            &
   itypps , ifacel , ifabor , ifmfbr , ifmcel , iprfml ,          &
   ipnfac , nodfac , ipnfbr , nodfbr ,                            &
   lstcel , lstfac , lstfbr ,                                     &
   idevel , ituser , ia     ,                                     &
   xyzcen , surfac , surfbo , cdgfac , cdgfbo , xyznod , volume , &
   dt     , rtpa   , rtp    , propce , propfa , propfb ,          &
   coefa  , coefb  , statis ,                                     &
   tracel , trafac , trafbr , rdevel , rtuser , ra     )

!===============================================================================
! Purpose:
! -------

!    User subroutine.

!    Output additional variables on a postprocessing mesh.

! Several "automatic" postprocessing meshes may be defined:
! - The volume mesh (ipart=-1) if 'ichrvl' = 1
! - The boundary mesh (ipart=-2) if 'ichrbo' = 1
! - SYRTHES coupling surface (ipart < -2) if 'ichrsy' = 1
! - Cooling tower exchange zone meshes (ipart < -2) if 'ichrze' = 1
!
! Additional meshes (cells or faces) may also be defined using the
! 'usdpst' user subroutine, (and possibly modified using 'usmpst').

! This subroutine is called once for each post-processing mesh
! (with a different value of 'ipart') for each time step at which output
! on this mesh is active.

!-------------------------------------------------------------------------------
! Arguments
!__________________.____._____.________________________________________________.
! name             !type!mode ! role                                           !
!__________________!____!_____!________________________________________________!
! idbia0           ! i  ! <-- ! number of first free position in ia            !
! idbra0           ! i  ! <-- ! number of first free position in ra            !
! ipart            ! i  ! <-- ! number of the post-processing mesh (< 0 or > 0)!
! ndim             ! i  ! <-- ! spatial dimension                              !
! ncelet           ! i  ! <-- ! number of extended (real + ghost) cells        !
! ncel             ! i  ! <-- ! number of cells                                !
! nfac             ! i  ! <-- ! number of interior faces                       !
! nfabor           ! i  ! <-- ! number of boundary faces                       !
! nfml             ! i  ! <-- ! number of families (group classes)             !
! nprfml           ! i  ! <-- ! number of properties per family (group class)  !
! nnod             ! i  ! <-- ! number of vertices                             !
! lndfac           ! i  ! <-- ! size of nodfac indexed array                   !
! lndfbr           ! i  ! <-- ! size of nodfbr indexed array                   !
! ncelbr           ! i  ! <-- ! number of cells with faces on boundary         !
! nvar             ! i  ! <-- ! total number of variables                      !
! nscal            ! i  ! <-- ! total number of scalars                        !
! nphas            ! i  ! <-- ! number of phases                               !
! nvlsta           ! i  ! <-- ! number of Lagrangian statistical variables     !
! ncelps           ! i  ! <-- ! number of cells in post-processing mesh        !
! nfacps           ! i  ! <-- ! number of interior faces in post-process. mesh !
! nfbrps           ! i  ! <-- ! number of boundary faces in post-process. mesh !
! nideve, nrdeve   ! i  ! <-- ! sizes of idevel and rdevel arrays              !
! nituse, nrtuse   ! i  ! <-- ! sizes of ituser and rtuser arrays              !
! itypps(3)        ! ia ! <-- ! global presence flag (0 or 1) for cells (1),   !
!                  !    !     ! interior faces (2), or boundary faces (3) in   !
!                  !    !     ! post-processing mesh                           !
! ifacel(2, nfac)  ! ia ! <-- ! interior faces -> cells connectivity           !
! ifabor(nfabor)   ! ia ! <-- ! boundary faces -> cells connectivity           !
! ifmfbr(nfabor)   ! ia ! <-- ! boundary face family numbers                   !
! ifmcel(ncelet)   ! ia ! <-- ! cell family numbers                            !
! iprfml           ! ia ! <-- ! property numbers per family                    !
!  (nfml, nprfml)  !    !     !                                                !
! ipnfac(nfac+1)   ! ia ! <-- ! interior faces -> vertices index (optional)    !
! nodfac(lndfac)   ! ia ! <-- ! interior faces -> vertices list (optional)     !
! ipnfbr(nfabor+1) ! ia ! <-- ! boundary faces -> vertices index (optional)    !
! nodfbr(lndfbr)   ! ia ! <-- ! boundary faces -> vertices list (optional)     !
! lstcel(ncelps)   ! ia ! <-- ! list of cells in post-processing mesh          !
! lstfac(nfacps)   ! ia ! <-- ! list of interior faces in post-processing mesh !
! lstfbr(nfbrps)   ! ia ! <-- ! list of boundary faces in post-processing mesh !
! idevel(nideve)   ! ia ! <-> ! integer work array for temporary development   !
! ituser(nituse)   ! ia ! <-> ! user-reserved integer work array               !
! ia(*)            ! ia ! --- ! main integer work array                        !
! xyzcen           ! ra ! <-- ! cell centers                                   !
!  (ndim, ncelet)  !    !     !                                                !
! surfac           ! ra ! <-- ! interior faces surface vectors                 !
!  (ndim, nfac)    !    !     !                                                !
! surfbo           ! ra ! <-- ! boundary faces surface vectors                 !
!  (ndim, nfabor)  !    !     !                                                !
! cdgfac           ! ra ! <-- ! interior faces centers of gravity              !
!  (ndim, nfac)    !    !     !                                                !
! cdgfbo           ! ra ! <-- ! boundary faces centers of gravity              !
!  (ndim, nfabor)  !    !     !                                                !
! xyznod           ! ra ! <-- ! vertex coordinates (optional)                  !
!  (ndim, nnod)    !    !     !                                                !
! volume(ncelet)   ! ra ! <-- ! cell volumes                                   !
! dt(ncelet)       ! ra ! <-- ! time step (per cell)                           !
! rtp, rtpa        ! ra ! <-- ! calculated variables at cell centers           !
!  (ncelet, *)     !    !     !  (at current and previous time steps)          !
! propce(ncelet, *)! ra ! <-- ! physical properties at cell centers            !
! propfa(nfac, *)  ! ra ! <-- ! physical properties at interior face centers   !
! propfb(nfabor, *)! ra ! <-- ! physical properties at boundary face centers   !
! coefa, coefb     ! ra ! <-- ! boundary conditions                            !
!  (nfabor, *)     !    !     !                                                !
! statis           ! ra ! <-- ! statistic values (Lagrangian)                  !
!  (ncelet, nvlsta)!    !     !                                                !
! tracel(*)        ! ra ! --- ! work array for post-processed cell values      !
! trafac(*)        ! ra ! --- ! work array for post-processed face values      !
! trafbr(*)        ! ra ! --- ! work array for post-processed boundary face v. !
! rdevel(nrdeve)   ! ra ! <-> ! real work array for temporary development      !
! rtuser(nrtuse)   ! ra ! <-> ! user-reserved real work array                  !
! ra(*)            ! ra ! --- ! main real work array                           !
!__________________!____!_____!________________________________________________!

!     Type: i (integer), r (real), s (string), a (array), l (logical),
!           and composite types (ex: ra real array)
!     mode: <-- input, --> output, <-> modifies data, --- work array
!===============================================================================

implicit none

!===============================================================================
! Common blocks
!===============================================================================

include "paramx.h"
include "cstnum.h"
include "pointe.h"
include "entsor.h"
include "optcal.h"
include "numvar.h"
include "parall.h"
include "period.h"

!===============================================================================

! Arguments

integer          idbia0, idbra0
integer          ipart
integer          ndim,   ncelet, ncel,   nfac,   nfabor
integer          nfml,   nprfml
integer          nnod,   lndfac, lndfbr, ncelbr
integer          nvar,   nscal , nphas , nvlsta
integer          ncelps, nfacps, nfbrps
integer          nideve, nrdeve, nituse, nrtuse

integer          itypps(3)
integer          ifacel(2,nfac), ifabor(nfabor)
integer          ifmfbr(nfabor), ifmcel(ncelet)
integer          iprfml(nfml,nprfml)
integer          ipnfac(nfac+1), nodfac(lndfac)
integer          ipnfbr(nfabor+1), nodfbr(lndfbr)
integer          lstcel(ncelps), lstfac(nfacps), lstfbr(nfbrps)
integer          idevel(nideve), ituser(nituse), ia(*)

double precision xyzcen(ndim,ncelet)
double precision surfac(ndim,nfac), surfbo(ndim,nfabor)
double precision cdgfac(ndim,nfac), cdgfbo(ndim,nfabor)
double precision xyznod(ndim,nnod), volume(ncelet)
double precision dt(ncelet), rtpa(ncelet,*), rtp(ncelet,*)
double precision propce(ncelet,*)
double precision propfa(nfac,*), propfb(nfabor,*)
double precision coefa(nfabor,*), coefb(nfabor,*)
double precision statis(ncelet,nvlsta)
double precision tracel(ncelps*3)
double precision trafac(nfacps*3), trafbr(nfbrps*3)
double precision rdevel(nrdeve), rtuser(nrtuse)
double precision ra(*)

! Local variables

character*32     namevr

integer          ntindp
integer          iel, ifac, iloc, iphas, ivar, iclt
integer          idimt, ii , jj
integer          idimte, itenso, ientla, ivarpr
integer          imom1, imom2, ipcmo1, ipcmo2, idtcm
double precision pond
double precision rvoid(1)

! For gradient and laplacien

integer          idebia, idebra
integer          ivar0 , inc   , iccocg
integer          iphydp, nswrgp, imligp, iwarnp, iclimv
integer          ipcsii
double precision epsrgp, climgp, extrap
double precision v1(ncelet,3),v2(ncelet,3)

! For laplacien
double precision diff(ncelet)
double precision viscf(nfac)  , viscb(nfabor)
integer          idtvarp,iconvp,idiffp,ircflp,ischcp,isstpp,imrgrap,ipp
double precision blencp,relaxp,thetap
double precision dpdx(ncelet)   , dpdy(ncelet)   , dpdz(ncelet)
double precision dpdxa(ncelet)  , dpdya(ncelet)  , dpdza(ncelet) 
double precision w1(ncelet)
double precision coefaf(nfabor), coefbf(nfabor)

integer          ipass
data             ipass /0/
save             ipass

!===============================================================================
! Analytical solution of Scalar (for check)

    idimt  = 1
    NAMEVR = 'Scalar_CHECK'

    ivar = isca(1)

    do iloc = 1, ncelps

      iel = lstcel(iloc)

      tracel(iloc) = xyzcen(3,iel)**3

    enddo

    ientla = 0
    ivarpr = 0

    call psteva(ipart, namevr, idimt, ientla, ivarpr,           &
    !==========
                ntcabs, ttcabs, tracel, trafac, trafbr)

!===============================================================================
! Difference between analytical solution of Scalar and real scalar (for check)

    idimt  = 1
    NAMEVR = 'Scalar_DIFF'

    ivar = isca(1)

    do iloc = 1, ncelps

      iel = lstcel(iloc)

      tracel(iloc) = abs(rtp(iel,isca(1)) - xyzcen(3,iel)**3)

    enddo

    ientla = 0
    ivarpr = 0

    call psteva(ipart, namevr, idimt, ientla, ivarpr,           &
    !==========
                ntcabs, ttcabs, tracel, trafac, trafbr)

!===============================================================================
! Analytical solution of Gradient (for check)

    idimt  = 1
    NAMEVR = 'Gradient_CHECK'

    ivar = isca(1)
    !iclimv = iclrtp(ivar,icoef)


     do iloc = 1, ncelps

      iel = lstcel(iloc)
      tracel(iloc) = 3.0d0*xyzcen(3,iel)**2

    enddo

    ientla = 0
    ivarpr = 0

    call psteva(ipart, namevr, idimt, ientla, ivarpr,           &
    !==========
                ntcabs, ttcabs, tracel, trafac, trafbr)

!===============================================================================
! Analytical solution of Laplacian (for check)

    idimt  = 1
    NAMEVR = 'Laplacian_CHECK'

    ivar = isca(1)
    !iclimv = iclrtp(ivar,icoef)


    do iloc = 1, ncelps

      iel = lstcel(iloc)

      tracel(iloc) = 6.0d0*xyzcen(3,iel) 

    enddo

    ientla = 0
    ivarpr = 0

    call psteva(ipart, namevr, idimt, ientla, ivarpr,           &
    !==========
                ntcabs, ttcabs, tracel, trafac, trafbr)


!===============================================================================
! Gradient with gradcel

    idimt  = 3
    NAMEVR = 'Gradient'

    ivar = isca(1)
    iclimv = iclrtp(ivar,icoef)

    inc = 1
    iccocg = 1
    nswrgp = nswrgr(ivar)
    imligp = imligr(ivar)
    iwarnp = iwarni(ivar)
    epsrgp = epsrgr(ivar)
    climgp = climgr(ivar)
    extrap = extrag(ivar)
    ivar0 = 0
    iphydp = 0

    do iel = 1, ncelet
     v1(iel,1) = 0.0d0
     v1(iel,2) = 0.0d0
     v1(iel,3) = 0.0d0
    enddo

    call grdcel                                                     &
   !==========
 ( idebia , idebra ,                                              &
   ndim   , ncelet , ncel   , nfac   , nfabor , nfml   , nprfml , &
   nnod   , lndfac , lndfbr , ncelbr , nphas  ,                   &
   nideve , nrdeve , nituse , nrtuse ,                            &
   ivar0  , imrgra , inc    , iccocg , nswrgp , imligp , iphydp , &
   iwarnp , nfecra , epsrgp , climgp , extrap ,                   &
   ifacel , ifabor , ifmfbr , ifmcel , iprfml ,                   &
   ipnfac , nodfac , ipnfbr , nodfbr ,                            &
   idevel , ituser , ia     ,                                     &
   xyzcen , surfac , surfbo , cdgfac , cdgfbo , xyznod , volume , &
   ra     , ra     , ra     ,                                     &
   rtp(1,ivar), coefa(1,iclimv) , coefb(1,iclimv)  ,              &
   v1(1,1) , v1(1,2) , v1(1,3) ,                      &
   v2(1,1) , v2(1,2) , v2(1,3) ,                                &
   rdevel , rtuser , ra     )

!
  ientla = 0
  ivarpr = 1

    call psteva(ipart, namevr, idimt, ientla, ivarpr,           &
    !==========
                ntcabs, ttcabs, v1, trafac, trafbr)

!===============================================================================
! Difference between analytical solution of Gradient and real Gradient (for check)

    idimt  = 1
    NAMEVR = 'Gradient_DIFF_z'

    ivar = isca(1)

    do iloc = 1, ncelps

      iel = lstcel(iloc)

      tracel(iloc) = abs(v1(iel,3) - 3.0d0*xyzcen(3,iel)**2)

    enddo

    ientla = 0
    ivarpr = 0

    call psteva(ipart, namevr, idimt, ientla, ivarpr,           &
    !==========
                ntcabs, ttcabs, tracel, trafac, trafbr)

!===============================================================================
! Laplacian ... avec bilsc2

   idimt  = 1
   NAMEVR = 'Laplacian'

   do iel = 1, ncelet
      diff(iel) = 1.0d0
   enddo

   call viscfa                                                   &
    !==========
 ( idbia0 , idbra0 ,                                              & 
   ndim   , ncelet , ncel   , nfac   , nfabor , nfml   , nprfml , & 
   nnod   , lndfac , lndfbr , ncelbr ,                            & 
   nideve , nrdeve , nituse , nrtuse , imvisf ,                   & 
   ifacel , ifabor , ifmfbr , ifmcel , iprfml ,                   & 
   ipnfac , nodfac , ipnfbr , nodfbr ,                            & 
   idevel , ituser , ia     ,                                     &
   xyzcen , surfac , surfbo , cdgfac , cdgfbo , xyznod , volume , & 
   diff ,                                                         & 
   viscf  , viscb  ,                                            & 
   rdevel , rtuser , ra     )   

! We have the diffusion
! So we can call bilsc2

    ivar = isca(1)
    iclimv = iclrtp(ivar,icoef)

! Autres definition

  idtvarp = 0 ! Pas de schema temporel
  iconvp = 0 ! Convection
  idiffp = 1 ! Diffusion
  nswrgp = 100 ! Nombre de sweep pour reconstruction gradient (100)
  imligp = -1 
  ircflp =  1
  ischcp =  1
  isstpp =  0
  inc    =  1
  imrgrap =  0
  iccocg =  1
  ipp    =  1
  ivar0  =  0
  iwarnp = iwarni(ivar)
  blencp = blencv(ivar)
  epsrgp = epsrgr(ivar)
  climgp = climgr(ivar)
  extrap = extrag(ivar)
  relaxp = relaxv(ivar)
  thetap = 1.0d0 !thetav(ivar) 


    do iel = 1, ncelet
     w1(iel) = 0.0d0
    enddo

    ! It's possibly not good.... !!!
    do ifac = 1, nfabor
     coefaf(ifac) = coefa(ifac,iclimv) !
     coefbf(ifac) = coefb(ifac,iclimv) !
    enddo


   call bilsc2                                                    &
   !================
 ( idbia0 , idbra0 ,                                              & 
   ndim   , ncelet , ncel   , nfac   , nfabor , nfml   , nprfml , & 
   nnod   , lndfac , lndfbr , ncelbr ,                            & 
   nvar   , nscal  , nphas  ,                                     & 
   nideve , nrdeve , nituse , nrtuse ,                            & 
   idtvarp , ivar0   , iconvp , idiffp , nswrgp , imligp , ircflp , & 
   ischcp , isstpp , inc    , imrgrap , iccocg ,                   & 
   ipp    , iwarnp ,                                              & 
   blencp , epsrgp , climgp , extrap , relaxp , thetap ,          & 
   ifacel , ifabor , ifmfbr , ifmcel , iprfml ,                   & 
   ipnfac , nodfac , ipnfbr , nodfbr ,                            & 
   idevel , ituser , ia     ,                                     & 
   xyzcen , surfac , surfbo , cdgfac , cdgfbo , xyznod , volume , & 
   rtp(1,ivar) , rtpa(1,ivar), coefa(1,iclimv), coefb(1,iclimv),coefaf, coefbf, &
   ra, ra,  viscf, viscb,                                &    
   w1,                                                            &
   dpdx   , dpdy   , dpdz   , dpdxa  , dpdya  , dpdza  ,          &
   rdevel , rtuser , ra     )         
                                    



    do iloc = 1, ncelps

      iel = lstcel(iloc)

      tracel(iloc) = w1(iel)/volume(iel)

    enddo

    ientla = 0
    ivarpr = 0

    call psteva(ipart, namevr, idimt, ientla, ivarpr,           &
    !==========
                ntcabs, ttcabs, tracel, trafac, trafbr)


!===============================================================================
! Difference with analytic solution

    idimt  = 1
    NAMEVR = 'Laplacian_DIFF'

    ivar = isca(1)


    do iloc = 1, ncelps

      iel = lstcel(iloc)

      tracel(iloc) = abs(tracel(iloc) - 6.0d0*xyzcen(3,iel))

    enddo

    ientla = 0
    ivarpr = 0

    call psteva(ipart, namevr, idimt, ientla, ivarpr,           &
    !==========
                ntcabs, ttcabs, tracel, trafac, trafbr)



return

end subroutine
